import os
from torch import Tensor
from tqdm.auto import tqdm

import joblib
from minio import Minio

import pandas as pd
import numpy as np
import scipy.sparse as sp

import torch
from torch.utils.data import TensorDataset
from transformers import AdamW

from examples.trainer.preprocessing import get_tfidf_features
from knodle.trainer import MajorityConfig, KNNConfig, SnorkelConfig, SnorkelKNNConfig
from knodle.model.logistic_regression_model import LogisticRegressionModel

# This python script contains rarely any explanation. For more description, we refer to the corresponding
# jupyter notebook. There the steps are explained in more detail


# Define some functions
from knodle.trainer.multi_trainer import MultiTrainer
from knodle.trainer.wscrossweigh.config import WSCrossWeighConfig


def np_array_to_tensor_dataset(x: np.ndarray) -> TensorDataset:
    if isinstance(x, sp.csr_matrix):
        x = x.toarray()
    x = torch.from_numpy(x)
    x = TensorDataset(x)
    return x


# Define constants
imdb_data_dir = os.path.join(os.getcwd(), "datasets", "spam")
processed_data_dir = os.path.join(imdb_data_dir, "processed")
os.makedirs(processed_data_dir, exist_ok=True)

# Download data
client = Minio("knodle.dm.univie.ac.at", secure=False)
files = [
    "df_train.csv", "df_test.csv",
    "train_rule_matches_z.lib", "test_rule_matches_z.lib",
    "mapping_rules_labels_t.lib"
]
for file in tqdm(files):
    client.fget_object(
        bucket_name="knodle",
        object_name=os.path.join("datasets/spam/processed", file),
        file_path=os.path.join(processed_data_dir, file),
    )

# Load data into memory
df_train = pd.read_csv(os.path.join(processed_data_dir, "df_train.csv"))
df_test = pd.read_csv(os.path.join(processed_data_dir, "df_test.csv"))

mapping_rules_labels_t = joblib.load(os.path.join(processed_data_dir, "mapping_rules_labels_t.lib"))

train_rule_matches_z = joblib.load(os.path.join(processed_data_dir, "train_rule_matches_z.lib"))
test_rule_matches_z = joblib.load(os.path.join(processed_data_dir, "test_rule_matches_z.lib"))

print(f"Train Z dimension: {train_rule_matches_z.shape}")
print(f"Train avg. matches per sample: {train_rule_matches_z.sum() / train_rule_matches_z.shape[0]}")

# tfidf
X_train_tfidf, X_test_tfidf, _ = get_tfidf_features(
    train_data=df_train["sample"].tolist(),
    test_data=df_test["sample"].tolist()
)
# convert input features to datasets
X_train_tfidf_dataset = TensorDataset(Tensor(X_train_tfidf.toarray()))
X_test_tfidf_dataset = TensorDataset(Tensor(X_test_tfidf.toarray()))

# get test labels
y_test = np_array_to_tensor_dataset(df_test['label'].values)

# initilize model
logreg_model = LogisticRegressionModel(X_train_tfidf.shape[1], 2)

configs = [
    MajorityConfig(optimizer=AdamW, lr=1e-4, batch_size=16, epochs=3),
    KNNConfig(optimizer=AdamW, k=2, lr=1e-4, batch_size=32, epochs=2),
    SnorkelConfig(optimizer=AdamW),
    SnorkelKNNConfig(optimizer=AdamW, radius=0.8),
    WSCrossWeighConfig(optimizer=AdamW)
]


trainer = MultiTrainer(
    name=["majority", "knn", "snorkel", "snorkel_knn", "wscrossweigh"],
    model=logreg_model,
    mapping_rules_labels_t=mapping_rules_labels_t,
    model_input_x=X_train_tfidf_dataset,
    rule_matches_z=train_rule_matches_z,
    trainer_config=configs,
)

# Run training
trainer.train()

# Run evaluation
metrics = trainer.test(X_test_tfidf_dataset, y_test)
for trainer, metric in metrics.items():
    print(f"Trainer: {trainer}, accuracy: {metric[0].get('accuracy')}")



import numpy as np


def transform_snorkel_matrix_to_z_t(class_matrix: np.ndarray) -> [np.ndarray, np.ndarray]:
    """Takes a matrix in format used by e.g. Snorkel (https://github.com/snorkel-team/snorkel)
    and transforms it to z / t matrices. Format
        - class_matrix_ij = -1, iff the rule doesn't apply
        - class_matrix_ij = k, iff the rule labels class k

    :param class_matrix: shape=(num_samples, num_weak_labellers)
    :return: Z matrix - binary encoded array of which rules matched. Shape: instances x rules.
             T matrix - mapping of rules to labels, binary encoded. Shape: rules x classes.
    """
    num_classes = class_matrix.max() + 1

    z_matrix = np.where(class_matrix != -1, 1, class_matrix)
    z_matrix = np.where(z_matrix == -1, 0, z_matrix)

    t_matrix = np.zeros((class_matrix.shape[1], num_classes))

    for i in range(class_matrix.shape[1]):
        row = class_matrix[:, i].tolist()
        label = max(row)
        if set(row) != {-1, label}:
            raise RuntimeError(
                "A weak labeller is not allowed to label more than one class."
            )

        t_matrix[i, label] = 1

    return z_matrix, t_matrix



from tests.trainer.generic import std_trainer_input_2

from knodle.trainer.wscrossweigh.wscrossweigh_weights_calculator import WSCrossWeighWeightsCalculator


def test_dscw_base_test(std_trainer_input_2):
    (
        model,
        inputs_x, mapping_rules_labels_t, train_rule_matches_z,
        test_dataset, test_labels
    ) = std_trainer_input_2

    trainer = WSCrossWeighWeightsCalculator(
        model=model,
        mapping_rules_labels_t=mapping_rules_labels_t,
        model_input_x=inputs_x,
        rule_matches_z=train_rule_matches_z
    )

    trainer.train()
    clf_report, _ = trainer.test(test_dataset, test_labels)

    # Check that this runs without error
    assert True



import numpy as np
import torch


class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""

    def __init__(self, patience=7, verbose=False, delta=0, name="checkpoint"):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        self.name = name

    def __call__(self, val_loss, model):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
        elif score < self.best_score + self.delta:
            self.counter += 1
            print(f"EarlyStopping counter: {self.counter} out of {self.patience}")
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            self.counter = 0

    def save_checkpoint(self, val_loss, model):
        """Saves model when validation loss decrease."""
        if self.verbose:
            print(
                f"Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ..."
            )
        torch.save(model.state_dict(), "models/" + self.name + ".pt")
        self.val_loss_min = val_loss



import os
import logging

import joblib
import numpy as np

from torch.optim import SGD
from torch.utils.data import TensorDataset

from scipy.sparse import csr_matrix
from sklearn.neighbors import NearestNeighbors
from annoy import AnnoyIndex

from knodle.transformation.majority import input_to_majority_vote_input
from knodle.transformation.torch_input import input_labels_to_tensordataset

from knodle.trainer.auto_trainer import AutoTrainer
from knodle.trainer.baseline.majority import MajorityVoteTrainer
from knodle.trainer.knn_denoising.config import KNNConfig
from knodle.trainer.utils.denoise import activate_neighbors

logger = logging.getLogger(__name__)


@AutoTrainer.register('knn')
class KnnDenoisingTrainer(MajorityVoteTrainer):
    def __init__(
            self,
            knn_feature_matrix: np.ndarray = None,
            **kwargs
    ):
        if kwargs.get("trainer_config") is None:
            kwargs["trainer_config"] = KNNConfig(optimizer=SGD, lr=0.001)
        super().__init__(**kwargs)

        if knn_feature_matrix is None:
            self.knn_feature_matrix = csr_matrix(self.model_input_x.tensors[0].numpy())
        else:
            self.knn_feature_matrix = knn_feature_matrix

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        self._load_train_params(model_input_x, rule_matches_z, dev_model_input_x, dev_gold_labels_y)

        # initialise optimizer
        self.trainer_config.optimizer = self.initialise_optimizer()

        self.rule_matches_z = self.rule_matches_z.astype(np.int8)
        self.mapping_rules_labels_t = self.mapping_rules_labels_t.astype(np.int8)

        self._knn_denoise_rule_matches()

        model_input_x, label_probs = input_to_majority_vote_input(
            self.model_input_x, self.rule_matches_z, self.mapping_rules_labels_t.astype(np.int64),
            filter_non_labelled=self.trainer_config.filter_non_labelled,
            other_class_id=self.trainer_config.other_class_id
        )

        feature_label_dataset = input_labels_to_tensordataset(model_input_x, label_probs)
        feature_label_dataloader = self._make_dataloader(feature_label_dataset)

        self._train_loop(feature_label_dataloader)

    def _knn_denoise_rule_matches(self) -> np.ndarray:
        """
        Denoises the applied weak supervision source.
        Args:
            rule_matches_z: Matrix with all applied weak supervision sources. Shape: (Instances x Rules)
        Returns: Denoised / Improved applied labeling function matrix. Shape: (Instances x Rules)
        """
        k = self.trainer_config.k
        if k == 1:
            return self.rule_matches_z

        # load cached data, if available
        if self.trainer_config.caching_folder:
            cache_file = self.trainer_config.get_cache_file()
            if os.path.isfile(cache_file):
                logger.info(f"Loaded knn matrix from cache: {cache_file}")
                return joblib.load(cache_file)

        logger.info(f"Start denoising labeling functions with k: {k}.")

        # ignore zero-match rows for knn construction & activation
        if self.trainer_config.activate_no_match_instances:
            ignore = np.zeros((self.knn_feature_matrix.shape[0],), dtype=np.bool)
        else:
            ignore = self.rule_matches_z.sum(-1) == 0

        # Set up data structure, to quickly find nearest neighbors
        if self.trainer_config.use_approximation:
            # use annoy fast ANN
            if k is not None:
                knn_matrix_shape = self.knn_feature_matrix.shape

                logger.info("Creating annoy index...")
                t = AnnoyIndex(knn_matrix_shape[1], 'dot')
                for i, v in enumerate(self.knn_feature_matrix):
                    if not ignore[i]:
                        t.add_item(i, v)

                t.build(10, n_jobs=self.trainer_config.n_jobs)

                self.knn_feature_matrix = None

                logger.info("Retrieving neighbor indices...")
                indices = (         # make a generator: no memory is allocated at this moment
                    np.array(t.get_nns_by_item(i, k, search_k=-1, include_distances=False))
                    if not ignore[i] else np.array([])
                    for i in range(knn_matrix_shape[0])
                )
            else:
                # possible radius implementation; delete error in config then
                pass
        else:
            # use standard precise kNN
            if k is not None:
                logger.info("Creating NN index...")
                neighbors = NearestNeighbors(n_neighbors=k, n_jobs=self.trainer_config.n_jobs)\
                    .fit(self.knn_feature_matrix)
                logger.info("Retrieving neighbor indices...")
                indices = neighbors.kneighbors(self.knn_feature_matrix, n_neighbors=k, return_distance=False)
            else:
                logger.info("Creating NN index...")
                neighbors = NearestNeighbors(radius=self.trainer_config.radius, n_jobs=self.trainer_config.n_jobs)\
                    .fit(self.knn_feature_matrix)
                logger.info("Retrieving neighbor indices...")
                indices = neighbors.radius_neighbors(self.knn_feature_matrix, return_distance=False)

        # activate matches.
        logger.info("Activating neighbors...")
        self.rule_matches_z = activate_neighbors(self.rule_matches_z, indices)

        # save data for caching
        if self.trainer_config.caching_folder:
            os.makedirs(self.trainer_config.caching_folder, exist_ok=True)
            joblib.dump(self.rule_matches_z, cache_file)

        return self.rule_matches_z

    def print_step_update(self, step: int, max_steps: int):
        if step % 40 == 0 and not step == 0:
            logger.info(f"  Batch {step}  of  {max_steps}.")



import logging

import numpy as np

import torch.nn as nn
from torch.optim import SGD
from torch.utils.data import TensorDataset

from knodle.transformation.majority import input_to_majority_vote_input
from knodle.transformation.torch_input import input_labels_to_tensordataset

from knodle.trainer.trainer import BaseTrainer
from knodle.trainer.auto_trainer import AutoTrainer
from knodle.trainer.baseline.config import MajorityConfig
from knodle.transformation.filter import filter_probability_threshold

logger = logging.getLogger(__name__)


@AutoTrainer.register('majority')
class MajorityVoteTrainer(BaseTrainer):
    """
    The baseline class implements a baseline model for labeling data with weak supervision.
        A simple majority vote is used for this purpose.
    """

    def __init__(self, **kwargs):
        if kwargs.get("trainer_config") is None:
            kwargs["trainer_config"] = MajorityConfig(optimizer=SGD, lr=0.001)
        super().__init__(**kwargs)

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        """
        This function gets final labels with a majority vote approach and trains the provided model.
        """
        self._load_train_params(model_input_x, rule_matches_z, dev_model_input_x, dev_gold_labels_y)
        self._apply_rule_reduction()

        # initialise optimizer
        self.trainer_config.optimizer = self.initialise_optimizer()

        self.model_input_x, noisy_y_train, self.rule_matches_z = input_to_majority_vote_input(
            self.rule_matches_z,
            self.mapping_rules_labels_t,
            self.model_input_x,
            use_probabilistic_labels=self.trainer_config.use_probabilistic_labels,
            filter_non_labelled=self.trainer_config.filter_non_labelled,
            probability_threshold=self.trainer_config.probability_threshold,
            other_class_id=self.trainer_config.other_class_id
        )

        feature_label_dataset = input_labels_to_tensordataset(self.model_input_x, noisy_y_train)
        feature_label_dataloader = self._make_dataloader(feature_label_dataset)

        self._train_loop(feature_label_dataloader)




import torch
from torch import nn


class LogisticRegressionModel(nn.Module):
    def __init__(self, input_dim: int, output_classes: int):
        super(LogisticRegressionModel, self).__init__()
        self.linear = torch.nn.Linear(input_dim, output_classes)

    def forward(self, x):
        x = x.float()
        outputs = self.linear(x)
        return outputs



import pytest

import numpy as np
import torch
from torch.utils.data import TensorDataset

from knodle.model.logistic_regression_model import LogisticRegressionModel


@pytest.fixture
def std_trainer_input_1():
    num_samples = 64
    num_features = 16
    num_rules = 6
    num_classes = 2

    x_np = np.ones((num_samples, num_features)).astype(np.float32)
    x_tensor = torch.from_numpy(x_np)
    model_input_x = TensorDataset(x_tensor)

    rule_matches_z = np.zeros((num_samples, num_rules))
    rule_matches_z[0, 0] = 1
    rule_matches_z[1:, 1] = 1

    mapping_rules_labels_t = np.zeros((num_rules, num_classes))
    mapping_rules_labels_t[:, 1] = 1

    y_np = np.ones((num_samples,))
    y_labels = TensorDataset(torch.from_numpy(y_np))

    model = LogisticRegressionModel(num_features, num_classes)

    return (
        model,
        model_input_x, rule_matches_z, mapping_rules_labels_t,
        y_labels
    )


@pytest.fixture
def std_trainer_input_2():
    model = LogisticRegressionModel(5, 2)

    inputs_x = TensorDataset(torch.Tensor(np.array([[1, 1, 1, 1, 1],
                                                    [2, 2, 2, 2, 2],
                                                    [3, 3, 3, 3, 3],
                                                    [6, 6, 6, 6, 6],
                                                    [7, 7, 7, 7, 7]])))

    mapping_rules_labels_t = np.array([[1, 0], [1, 0], [0, 1]])
    train_rule_matches_z = np.array([[1, 0, 0], [1, 1, 0], [1, 0, 1], [0, 1, 0], [0, 0, 1]])

    test_dataset = TensorDataset(torch.Tensor(np.array([[4, 4, 4, 4, 4], [5, 5, 5, 5, 5]])))
    test_labels = TensorDataset(torch.Tensor(np.array([0, 1])))

    return (
        model,
        inputs_x, mapping_rules_labels_t, train_rule_matches_z,
        test_dataset, test_labels
    )



import random
import logging

import torch
from torch import Tensor, argmax
from torch.utils.data import TensorDataset
import numpy as np
import matplotlib.pyplot as plt


def log_section(text: str, logger: logging, additional_info: {} = None) -> None:
    """
    Prints a section
    Args:
        text: Text to print
        logger: Logger object

    Returns:

    """
    logger.info("======================================")
    logger.info(text)
    if additional_info:
        for key, value in additional_info.items():
            logger.info("{}: {}".format(key, value))
    logger.info("======================================")


def accuracy_of_probs(predictions: Tensor, ground_truth: Tensor):
    """
    Function to calculate the accuracy of two tensors with probabilities attached.
    Args:
        predictions: Predictions, shape: instances x labels
        ground_truth: Ground truth, shape one of (instances x 1) or (instances x labels)

    Returns: Accuracy

    """
    ground_truth = (
        ground_truth if len(ground_truth.shape) == 1 else argmax(ground_truth, dim=-1)
    )
    y_pred_max = argmax(predictions, dim=-1)
    correct_pred = (y_pred_max == ground_truth).float()
    acc = correct_pred.sum() / len(correct_pred)
    return acc


def extract_tensor_from_dataset(dataset: TensorDataset, tensor_index: int) -> Tensor:
    """
    Extracts a tensor from a dataset.
    Args:
        dataset: Dataset to extract tensor from
        tensor_index: Which tensor to extract

    Returns: Tensor

    """
    return dataset.tensors[tensor_index]


def check_and_return_device() -> torch.device:
    """
    Function to check if a GPU is available and sets the device for pytorch.
    If a GPU is available -> The device is GPU, if not the device is CPU
    """
    if torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")
    return device


def set_seed(seed: int) -> None:
    """ Fix seed for all shuffle processes in order to get the reproducible result """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)



from typing import Tuple

import numpy as np
from snorkel.labeling.model import LabelModel

from torch.optim import SGD
from torch.utils.data import TensorDataset

from knodle.transformation.torch_input import input_labels_to_tensordataset

from knodle.trainer.auto_trainer import AutoTrainer
from knodle.trainer.baseline.majority import MajorityVoteTrainer
from knodle.trainer.knn_aggregation.knn import KNNAggregationTrainer

from knodle.trainer.snorkel.config import SnorkelConfig, SnorkelKNNConfig
from knodle.trainer.snorkel.utils import (
    z_t_matrix_to_snorkel_matrix,
    prepare_empty_rule_matches,
    add_labels_for_empty_examples
)
from knodle.transformation.filter import filter_tensor_dataset_by_indices


@AutoTrainer.register('snorkel')
class SnorkelTrainer(MajorityVoteTrainer):
    """Provides a wrapper around the Snorkel system. See https://github.com/snorkel-team/snorkel for more details.
    Formally, a generative model P(Y, Y') is learned, with Y' = Z * T, followed by a discriminative model specified
    by the user.
    """

    def __init__(self, **kwargs):
        if kwargs.get("trainer_config", None) is None:
            kwargs["trainer_config"] = SnorkelConfig(optimizer=SGD, lr=0.001)
        super().__init__(**kwargs)

    def _snorkel_denoising(
            self, model_input_x: TensorDataset, rule_matches_z: np.ndarray
    ) -> Tuple[TensorDataset, np.ndarray]:
        """
		Trains the generative model.
        Premise:
            Snorkel can not make use of rule-unlabeled examples (no rule matches).
            The generative LabelModel assigns such examples a uniform distribution over all available labels,
            which contradicts the desired behaviour. Such examples should be either filtered or assigned an
            "other class id".
        Filtering / other class strategy:
            filter_non_labelled = True:
                Drop the unlabeled examples completely prior to LabelModel.
            filter_non_labelled = False:
                However, we might want to keep negative examples in the training data, but we should not pass them
                through the LabelModel. Therefore, the rule-unlabeled part of the data skips the LabelModel step
                and is added directly to the output data with manually assigned "other class id".

        Args:
            model_input_x: feature input to the classifier
            rule_matches_z: input rule matches

        Returns:
            eventually filtered model input,
            corresponding probability distributions over labels generated by Snorkel. Shape: (#Instances x #Labels)
        """

        # initialise optimizer
        self.trainer_config.optimizer = self.initialise_optimizer()

        # create Snorkel matrix
        non_empty_mask, rule_matches_z = prepare_empty_rule_matches(rule_matches_z)
        L_train = z_t_matrix_to_snorkel_matrix(rule_matches_z, self.mapping_rules_labels_t)

        # train LabelModel
        label_model = LabelModel(cardinality=self.mapping_rules_labels_t.shape[1], verbose=True)
        fitting_kwargs = {}
        if self.trainer_config.seed is not None:
            fitting_kwargs["seed"] = self.trainer_config.seed

        label_model.fit(
            L_train,
            n_epochs=self.trainer_config.label_model_num_epochs,
            log_freq=self.trainer_config.label_model_log_freq,
            **fitting_kwargs
        )
        label_probs_gen = label_model.predict_proba(L_train)

        if self.trainer_config.filter_non_labelled:
            # filter out respective input irrevocably from all data
            model_input_x = filter_tensor_dataset_by_indices(dataset=model_input_x, filter_ids=non_empty_mask)
            label_probs = label_probs_gen
        else:
            # add "other class" labels for empty examples in model input x, that were not given to Snorkel
            label_probs = add_labels_for_empty_examples(
                label_probs_gen=label_probs_gen, non_zero_mask=non_empty_mask,
                output_classes=self.trainer_config.output_classes,
                other_class_id=self.trainer_config.other_class_id)
        return model_input_x, label_probs

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        self._load_train_params(model_input_x, rule_matches_z, dev_model_input_x, dev_gold_labels_y)
        self._apply_rule_reduction()

        model_input_x, label_probs = self._snorkel_denoising(self.model_input_x, self.rule_matches_z)

        # Standard training
        feature_label_dataset = input_labels_to_tensordataset(model_input_x, label_probs)
        feature_label_dataloader = self._make_dataloader(feature_label_dataset)

        self._train_loop(feature_label_dataloader)


@AutoTrainer.register('snorkel_knn')
class SnorkelKNNAggregationTrainer(SnorkelTrainer, KNNAggregationTrainer):
    """Calls k-NN denoising, before the Snorkel generative and discriminative training is started.
    """
    def __init__(self, **kwargs):
        if kwargs.get("trainer_config", None) is None:
            kwargs["trainer_config"] = SnorkelKNNConfig(optimizer=SGD, lr=0.001)
        super().__init__(**kwargs)

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        self._load_train_params(model_input_x, rule_matches_z, dev_model_input_x, dev_gold_labels_y)
        self._apply_rule_reduction()

        # Snorkel denoising
        denoised_rule_matches_z = self._knn_denoise_rule_matches()
        model_input_x, label_probs = self._snorkel_denoising(self.model_input_x, denoised_rule_matches_z)

        # Standard training
        feature_label_dataset = input_labels_to_tensordataset(model_input_x, label_probs)
        feature_label_dataloader = self._make_dataloader(feature_label_dataset)

        self._train_loop(feature_label_dataloader)



from typing import Callable

import numpy as np
from torch.utils.data import TensorDataset

from knodle.trainer.trainer import Trainer


class AutoTrainer:
    """ The factory class for creating training executors
    See See https://medium.com/@geoffreykoh/implementing-the-factory-
    pattern-via-dynamic-registry-and-python-decorators-479fc1537bbe
    """

    registry = {}
    """ Internal registry for available trainers """

    def __init__(self, name, **kwargs):
        self.trainer = self.create_trainer(name, **kwargs)

    def train(
            self,
            model_input_x: TensorDataset = None, rule_matches_z: np.ndarray = None,
            dev_model_input_x: TensorDataset = None, dev_gold_labels_y: TensorDataset = None
    ):
        return self.trainer.train(model_input_x, rule_matches_z, dev_model_input_x, dev_gold_labels_y)

    def test(self, test_features: TensorDataset, test_labels: TensorDataset):
        return self.trainer.test(test_features, test_labels)

    @classmethod
    def create_trainer(cls, name: str, **kwargs) -> Trainer:
        """ Factory command to create the executor """

        exec_class = cls.registry[name]
        executor = exec_class(**kwargs)
        return executor

    @classmethod
    def register(cls, name: str) -> Callable:
        def inner_wrapper(wrapped_class: Trainer) -> Callable:
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper



import torch
from torch import nn
import numpy as np


class LogisticRegressionModel(nn.Module):
    def __init__(
            self,
            input_size: int,
            word_input_dim: int,
            word_output_dim: int,
            word_embedding_matrix: np.ndarray,
            output_classes: int,
    ):
        super(LogisticRegressionModel, self).__init__()

        self.word_embedding = nn.Embedding(
            word_input_dim, word_output_dim, padding_idx=0
        )
        self.word_embedding.weight = nn.Parameter(
            torch.tensor(word_embedding_matrix, dtype=torch.float32)
        )
        self.word_embedding.weight.requires_grad = False

        # self.td_dense = nn.Linear(input_size * word_output_dim, size_factor)
        self.linear = nn.Linear(input_size * word_output_dim, output_classes)

    def forward(self, x):
        word_embeddings = self.word_embedding(x)
        word_embeddings = word_embeddings.view(x.shape[0], -1)
        outputs = self.linear(word_embeddings)
        return outputs



from typing import List, Union, Tuple
from joblib import dump

from sklearn.feature_extraction.text import TfidfVectorizer

import numpy as np
from torch.utils.data import TensorDataset


def get_tfidf_features(
        train_data: List, test_data: List = None, dev_data: List = None, path_to_cache: str = None,
        max_features: int = None
) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, None]]:
    """
    Convert input data to a matrix of TF-IDF features.
    :param train_data: training samples that are to be encoded with TF-IDF features. Can be given as Series or
    as DataFrames with specified column number where the sample are stored.
    :param test_data: if DataFrame/Series with test data is provided
    :param dev_data: if DataFrame/Series with development data is provided, it will be encoded as well
    :param path_to_cache: a path to the folder where calculated cached TF-IDF values should be saved
    :param max_features: If not None, build a vocabulary that only consider the top max_features ordered by term
    frequency across the corpus.
    :return: TensorDataset with encoded data
    """

    dev_transformed_data, test_transformed_data = None, None
    vectorizer = TfidfVectorizer(max_features=max_features)

    train_transformed_data = vectorizer.fit_transform(train_data)
    if test_data is not None:
        test_transformed_data = vectorizer.transform(test_data)
    if dev_data is not None:
        dev_transformed_data = vectorizer.transform(dev_data)

    if path_to_cache:
        dump(train_transformed_data, path_to_cache)
        dump(dev_transformed_data, path_to_cache)
        dump(test_transformed_data, path_to_cache)

    return train_transformed_data, test_transformed_data, dev_transformed_data


def convert_text_to_transformer_input(tokenizer, texts: List[str]) -> TensorDataset:
    """
    Convert input data to BERT encoded features (more details could be found at
    https://huggingface.co/transformers/model_doc)
    :param texts: training/dev/test samples that are to be encoded with BERT features. Can be given as Series or
    as DataFrames with specified column number where the sample are stored.
    :param tokenizer: DistilBertTokenizer tokenizer for english from HuggingFace
    :return: TensorDataset with encoded data
    """
    encoding = tokenizer(texts, return_tensors="pt", padding=True, truncation=True)
    input_ids = encoding.get('input_ids')
    attention_mask = encoding.get('attention_mask')

    input_values_x = TensorDataset(input_ids, attention_mask)

    return input_values_x



import logging
from typing import Iterable
from tqdm import tqdm

import numpy as np
import scipy.sparse as ss

logger = logging.getLogger(__name__)


def activate_neighbors(
        rule_matches_z: np.ndarray, indices: Iterable[np.ndarray]
) -> np.ndarray:
    """
    Take provided closest neighbors and add their rule matches.
    If no indices provided for an instance, no activation is done (no matches added).
    Args:
        rule_matches_z: All rule matches. Shape: instances x rules
        indices: Neighbor indices from knn per instance.
    Returns:
    """
    if isinstance(rule_matches_z, ss.csr_matrix):
        new_z_matrix = ss.lil_matrix(rule_matches_z.shape, dtype=np.int8)
    else:
        new_z_matrix = np.zeros_like(rule_matches_z)
    
    # make sure initial matches are preserved
    new_z_matrix[rule_matches_z != 0] = 1

    for index, neighbors in enumerate(tqdm(indices)):
        if len(neighbors) == 0:
            continue
            
        neighbors = neighbors.astype(np.int64)
        neighborhood_z = rule_matches_z[neighbors, :].astype(np.int8)

        activated_lfs = neighborhood_z.sum(axis=0)  # Add all lf activations

        # All with != 0 are valid. We could also do some division here for weighted
        activated_lfs = np.where(activated_lfs > 0)[0]

        # add neighbor matches to the target object
        new_z_matrix[index, activated_lfs] = 1
        
    # check out the amount of new matches
    logger.info(f"Match increase: {new_z_matrix.sum() / rule_matches_z.sum() * 100 - 100:.2f}%")
    
    if isinstance(new_z_matrix, ss.lil_matrix):
        return new_z_matrix.tocsr()
    else:
        return new_z_matrix



import itertools
import json
import logging
import re
from typing import Union, Tuple

import spacy
import pandas as pd
from pandas import DataFrame

ARG1 = "$ARG1"
ARG2 = "$ARG2"
PRINT_EVERY = 10000
UNKNOWN_RELATIONS_ID = 404  # id which will be assigned to unknown relations, i.e. relations that wasn't seen in KB

logger = logging.getLogger(__name__)


def get_analysed_conll_data(
        conll_data: str,
        patterns2regex: Union[dict, None],
        labels2ids: dict,
        perform_search: bool = False,
) -> Tuple[DataFrame, DataFrame]:
    """
    Reads conll data, extract information about sentences and gold labels. The sample are analysed with SpaCy package
    :param labels2ids: dictionary with labels and their ids
    :param conll_data: path to data saved in conll format
    :param patterns2regex: dictionary with pattern and their corresponding regexes
    :param perform_search: boolean whether indicates whether also pattern search in sentences is to be performed
    :return: DataFrame with fields "sample" (raw text samples), "gold_labels" (label that samples got in original
    conll set}) and "retrieved patterns" (matched patterns in samples; empty if pattern search wasn't performed)
    """
    all_lines = count_file_lines(conll_data)
    processed_lines = 0
    analyzer = spacy.load("en_core_web_sm")

    (
        samples,
        labels,
        enc_labels,
        patterns_retr,
        raw_patterns_retr,
        neg_samples,
        neg_labels,
        neg_enc_labels,
    ) = ([] for _ in range(8))

    with open(conll_data, encoding="utf-8") as f:
        for line in f:
            processed_lines += 1
            line = line.strip()
            if line.startswith("# id="):  # Instance starts
                sample = ""
                label = line.split(" ")[3][5:]
                enc_label = encode_labels(label, labels2ids)
            elif line == "":  # Instance ends
                if (
                        label == "no_relation" or enc_label == UNKNOWN_RELATIONS_ID
                ):  # skip no_relation samples or unknown rel
                    continue
                sample_spacy = analyzer(sample).to_json()
                sample_extractions = get_extracted_sample(sample_spacy)

                if perform_search:
                    (
                        raw_pattern_retrieved,
                        encoded_pattern_retrieved,
                    ) = retrieve_patterns_in_sample(sample_extractions, patterns2regex)
                    if encoded_pattern_retrieved:
                        samples.append(sample)
                        labels.append(label)
                        enc_labels.append(enc_label)
                        patterns_retr.append(encoded_pattern_retrieved)
                        raw_patterns_retr.append(raw_pattern_retrieved)
                    else:  # if nothing is found, add this sample preserving its original label
                        neg_samples.append(sample)
                        neg_labels.append(label)
                        neg_enc_labels.append(enc_label)
                else:
                    samples.append(sample)
                    labels.append(label)
                    enc_labels.append(enc_label)
                    patterns_retr.append([])
                    raw_patterns_retr.append([])

            elif line.startswith("#"):  # comment
                continue
            else:
                parts = line.split("\t")
                token = parts[1]
                if token == "-LRB-":
                    token = "("
                elif token == "-RRB-":
                    token = ")"
                sample += " " + token
            if processed_lines % PRINT_EVERY == 0:
                logger.info(
                    "Processed {:0.2f}% of {} file".format(
                        100 * processed_lines / all_lines, conll_data.split("/")[-1]
                    )
                )
    return build_df(
        samples,
        patterns_retr,
        raw_patterns_retr,
        labels,
        enc_labels,
        neg_samples,
        neg_labels,
        neg_enc_labels,
    )


def encode_labels(label: str, label2id: dict, other_class: int) -> int:
    """ Encodes labels with corresponding labels id. If relation is unknown, returns special id for unknown relations"""
    return label2id.get(label, other_class)


def build_df(
        samples: list,
        retrieved_patterns: list,
        raw_patterns_retrieved: list,
        labels: list,
        enc_labels: list,
        neg_samples: list,
        neg_labels: list,
        neg_enc_labels: list,
) -> (pd.DataFrame, pd.DataFrame):
    """
    This function builds two dataframes: one with samples where some rule matched (columns: sample, matches pattern,
    original label) and the second with labels that got no rule matches (columns: sample, label=original label)
    """
    samples = pd.DataFrame.from_dict(
        {
            "sample": samples,
            "retrieved_patterns": retrieved_patterns,
            "raw_retrieved_patterns": raw_patterns_retrieved,
            "labels": labels,
            "enc_labels": enc_labels,
        }
    )
    no_pattern_samples = pd.DataFrame.from_dict(
        {"sample": neg_samples, "enc_labels": neg_enc_labels, "labels": neg_labels}
    )
    return samples, no_pattern_samples


def get_id(item: Union[int, str], dic: dict) -> Union[int, str]:
    """This function checks if there is a key in a dict. If so, it returns a value of a dict.
    Creates a dictionary pair {key : max(dict value) + 1} otherwise"""
    if item in dic:
        item_id = dic[item]
    else:
        item_id = len(dic)
        dic[item] = item_id
    return item_id


def update_dict(key: Union[int, str], value: Union[int, str], dic: dict) -> None:
    """This function checks if there is a key in a dict. If so, it added new value to the list of values for this key.
    Create a {key: {value}} pair otherwise."""
    if key in dic:
        dic[key].append(value)
    else:
        dic[key] = [value]


def convert_pattern_to_regex(pattern: str) -> str:
    """ Transform the pattern into regex by char escaping and adding of optional articles in front of the arguments"""
    return re.sub("\\\\\\$ARG", "(A )?(a )?(The )?(the )?\\$ARG", re.escape(pattern))


def get_match_matrix_row(num_columns: int, matched_columns: list) -> list:
    """
    There is a function that creates a list which will be transformed into matrix row in future. All elements in the
    list equal zeros, apart from the ones with indices given in matched_columns list - they equal ones.
    :param num_columns: length of created list = future row vector 2nd dimensionality
    :param matched_columns: list of element indices that equal to 1
    :return: binary list
    """
    matrix_row = [0] * num_columns
    if matched_columns:
        for m_column in matched_columns:
            matrix_row[m_column] = 1
    return matrix_row


def get_extracted_sample(sample: dict) -> list:
    """
    Using the Spacy entity recognition information, forms the subtractions of the samples in form of
    "ARG1 <some words> ARG2" strings which will be used to look for patterns in them
    :param sample: sample analysed with SpaCy package and saved as dictionary
    :return: list of sample substrings
    """
    return [
        (ARG1 + sample["text"][ent1["end"]: ent2["start"]] + ARG2)
        if ent1["end"] < ent2["end"]
        else (ARG2 + sample["text"][ent2["end"]: ent1["start"]] + ARG1)
        for ent1, ent2 in itertools.permutations(sample["ents"], 2)
    ]


def retrieve_patterns_in_sample(
        extr_samples: list, pattern2regex: dict
) -> Union[Tuple[None, None], Tuple[list, list]]:
    """
    Looks for pattern in a sample and returns a list which would be turned into a row of a Z matrix.
    :param extr_samples: list of sample substrings in the form of "ARG1 <some words> ARG2", in which the patterns
    will be searching
    :param pattern2regex: dictionary with pattern and their corresponding regexes
    :return: if there was found smth, returns a row of a Z matrix as a list, where all elements equal 0 apart from
    the element with the index corresponding to the matched pattern id - this element equals 1.
    If no pattern matched, returns None.
    """
    matched_patterns = []
    for sample, pattern in itertools.product(extr_samples, pattern2regex):
        if re.search(pattern2regex[pattern], sample):
            matched_patterns.append(pattern)
    if len(matched_patterns) > 0:
        return matched_patterns, get_match_matrix_row(
            len(pattern2regex), list(set(matched_patterns))
        )
    return None, None


def count_file_lines(file_name: str) -> int:
    """ Count the number of line in a file """
    with open(file_name) as f:
        return len(f.readlines())


def save_dict(dict_: dict, path: str) -> None:
    with open(path, "w+") as f:
        json.dump(dict_, f)


def convert_to_tacred_rel(label: str) -> str:
    """ Function changes some labels in train data that are transcribed differently to tacred labels"""
    if label == "org:top_members_employees":
        return "org:top_members/employees"
    elif label == "org:political_religious_affiliation":
        return "org:political/religious_affiliation"
    elif label == "per:employee_or_member_of":
        return "per:employee_of"
    elif label == "org:number_of_employees_members":
        return "org:number_of_employees/members"
    elif label == "per:statesorprovinces_of_residence":
        return "per:stateorprovinces_of_residence"
    elif label == "org:date_founded":
        return "org:founded"
    elif label == "org:date_dissolved":
        return "org:dissolved"
    else:
        return label



import numpy as np
from scipy import sparse as ss

import torch
from torch.utils.data import TensorDataset
from knodle.trainer.snorkel.utils import (
    z_t_matrix_to_snorkel_matrix,
    prepare_empty_rule_matches,
    add_labels_for_empty_examples
)


def test_z_t_matrix_to_snorkel_matrix():
    # test dense case
    z = np.array([
        [0, 1, 0, 0],
        [0, 0, 1, 1]
    ])

    t = np.array([
        [1, 0],
        [0, 1],
        [1, 0],
        [0, 1]
    ])

    snorkel_gold = np.array([
        [-1, 1, -1, -1],
        [-1, -1, 0,  1]
    ])

    snorkel_test = z_t_matrix_to_snorkel_matrix(z, t)
    np.testing.assert_equal(snorkel_gold, snorkel_test)

    # test sparse case
    z = ss.csr_matrix([
        [0, 1, 0, 0],
        [0, 0, 1, 1]
    ])

    t = ss.csr_matrix([
        [1, 0],
        [0, 1],
        [1, 0],
        [0, 1]
    ])

    snorkel_gold = np.array([
        [-1, 1, -1, -1],
        [-1, -1, 0,  1]
    ])

    snorkel_test = z_t_matrix_to_snorkel_matrix(z, t)
    np.testing.assert_equal(snorkel_gold, snorkel_test)


def test_label_model_data():
    num_samples = 5
    num_rules = 6

    rule_matches_z = np.ones((num_samples, num_rules))
    rule_matches_z[[1, 4]] = 0

    non_zero_mask, out_rule_matches_z = prepare_empty_rule_matches(rule_matches_z)

    expected_mask = np.array([True, False, True, True, False])
    expected_rule_matches = np.ones((3, num_rules))

    np.testing.assert_equal(non_zero_mask, expected_mask)
    np.testing.assert_equal(out_rule_matches_z, expected_rule_matches)


def test_other_class_labels():
    label_probs_gen = np.array([
        [0.3, 0.6, 0.0, 0.1],
        [0.2, 0.2, 0.2, 0.4],
        [1.0, 0.0, 0.0, 0.0]
    ])
    output_classes = 5
    other_class_id = 4

    # test without empty rows
    non_zero_mask = np.array([True, True, True])
    expected_probs = np.array([
        [0.3, 0.6, 0.0, 0.1, 0.0],
        [0.2, 0.2, 0.2, 0.4, 0.0],
        [1.0, 0.0, 0.0, 0.0, 0.0]
    ])
    label_probs = add_labels_for_empty_examples(label_probs_gen, non_zero_mask, output_classes, other_class_id)

    np.testing.assert_equal(label_probs, expected_probs)

    # test with empty rows
    non_zero_mask = np.array([True, False, False, True, True])
    expected_probs = np.array([
        [0.3, 0.6, 0.0, 0.1, 0.0],
        [0.0, 0.0, 0.0, 0.0, 1.0],
        [0.0, 0.0, 0.0, 0.0, 1.0],
        [0.2, 0.2, 0.2, 0.4, 0.0],
        [1.0, 0.0, 0.0, 0.0, 0.0]
    ])
    label_probs = add_labels_for_empty_examples(label_probs_gen, non_zero_mask, output_classes, other_class_id)

    np.testing.assert_equal(label_probs, expected_probs)



import io
import os
import re
from typing import Dict

from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


VERSION: Dict[str, str] = {}
with open("knodle/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

with open("requirements.txt") as f:
    requirements = f.readlines()

test_requirements = ["pytest", "pytest-cov"]

# use external package to extract tag of the latest release as version
# https://pypi.org/project/setuptools-git-versioning/

setup(
    name="knodle",
    version=VERSION.get("__version__"),
    url="http://knodle.cc",
    project_urls={
        "github": "https://github.com/knodle/knodle",
        "Bug Tracker": "https://github.com/knodle/knodle/issues"
    },
    license="Apache 2.0",
    author="knodle",
    author_email="knodle@cs.univie.ac.at",
    description="Knowledge infused deep learning framework",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests"]),
    package_dir={'knodle': 'knodle'},
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require={"test": test_requirements},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
)



from tests.trainer.generic import std_trainer_input_1

from knodle.trainer.auto_trainer import AutoTrainer


def test_auto_train(std_trainer_input_1):
    (
        model,
        model_input_x, rule_matches_z, mapping_rules_labels_t,
        y_labels
    ) = std_trainer_input_1

    for name in AutoTrainer.registry:
        if name in ["majority", "snorkel"]:
            trainer = AutoTrainer(
                name=name,
                model=model,
                mapping_rules_labels_t=mapping_rules_labels_t,
                model_input_x=model_input_x,
                rule_matches_z=rule_matches_z,
            )

            trainer.train()
            metrics = trainer.test(model_input_x, y_labels)

            # Check whether the code ran up to here
            assert 2 == 2



from typing import Tuple

import numpy as np
from scipy import sparse as ss


def z_t_matrix_to_snorkel_matrix(rule_matches_z: np.ndarray, mapping_rules_labels_t: np.ndarray) -> np.ndarray:
    snorkel_matrix = -1 * np.ones(rule_matches_z.shape)

    if isinstance(rule_matches_z, ss.csr_matrix):
        rule_matches_z = rule_matches_z.toarray()

    z_to_t = np.argmax(mapping_rules_labels_t, axis=-1)
    print(z_to_t, type(z_to_t))
    if isinstance(mapping_rules_labels_t, ss.csr_matrix):
        # transform np.matrix to np.array
        z_to_t = np.array(z_to_t).flatten()

    for i in range(rule_matches_z.shape[0]):
        non_zero_idx = np.nonzero(rule_matches_z[i])[0]
        snorkel_matrix[i, non_zero_idx] = z_to_t[non_zero_idx]

    return snorkel_matrix


def prepare_empty_rule_matches(rule_matches_z: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Remove empty rows in rule matches for LabelModel.
    Args:
        rule_matches_z: input rule matches
    Returns:
        boolean mask indicating non-empty rows and filtered rule matches
    """
    # find empty rows of the rule matches
    non_zero_mask = rule_matches_z.sum(axis=1) != 0
    if isinstance(rule_matches_z, ss.csr_matrix):
        # transform np.matrix to np.array
        non_zero_mask = np.array(non_zero_mask).flatten()
    non_zero_indices = np.where(non_zero_mask)[0]

    # exclude empty rows from LabelModel input
    rule_matches_z = rule_matches_z[non_zero_indices]
    return non_zero_mask, rule_matches_z


def add_labels_for_empty_examples(
        label_probs_gen: np.ndarray, non_zero_mask: np.ndarray, output_classes: int, other_class_id: int
) -> np.ndarray:
    """
    Args:
        label_probs_gen: labels generated for the non-empty rows
        non_zero_mask:  boolean mask indicating which examples get the generated labels
        output_classes: number of output classes, with respect to the other_class
        other_class_id: id of the class for empty rows
    Returns: distribution of labels for both empty and non-empty rows (#instances x #classes)
    """
    # make dummy label distibutions for all of the examples
    # number of output classes is eventually t.shape[1]+1, if other class id should be added
    label_probs = np.zeros((non_zero_mask.shape[0], output_classes))

    # fill labels of the generative model
    label_probs[non_zero_mask, :label_probs_gen.shape[1]] = label_probs_gen

    # assign full probability to other class for empty rows
    label_probs[~non_zero_mask, other_class_id] = 1
    return label_probs



from knodle.trainer.baseline.config import MajorityConfig
from knodle.trainer.knn_aggregation.config import KNNConfig

from knodle.trainer.auto_config import AutoConfig


@AutoConfig.register("snorkel")
class SnorkelConfig(MajorityConfig):
    """Config class for the Snorkel wrapper.
    """

    def __init__(
            self,
            label_model_num_epochs: int = 5000,
            label_model_log_freq: int = 500,
            **kwargs
    ):
        """
        :param label_model_num_epochs: Number of epochs to train the Label model, which computes P(Y, Z, T).
        :param label_model_log_freq: Logging frequency.
        """
        super().__init__(**kwargs)
        self.label_model_num_epochs = label_model_num_epochs
        self.label_model_log_freq = label_model_log_freq


@AutoConfig.register("snorkel_knn")
class SnorkelKNNConfig(SnorkelConfig, KNNConfig):
    """Config class for the SnorkelkNNTrainer, which combines k-NN denoising and the Snorkel Label model, in that order.
    See base trainers for configuration possibilities.
    """

    def __init__(self, **kwargs):
        # use all config parameters in SnorkelConfig and KNNConfig
        super().__init__(**kwargs)



from knodle.trainer.baseline.config import MajorityConfig

from knodle.trainer.auto_config import AutoConfig


@AutoConfig.register("cleanlab")
class CleanLabConfig(MajorityConfig):
    def __init__(
            self,
            cv_n_folds=5,
            prune_method: str = 'prune_by_noise_rate',
            converge_latent_estimates=False,
            pulearning=None,
            n_jobs: int = None,
            psx_calculation_method: str = 'random',
            **kwargs
    ):
        """
        All CleanLab specific parameters are inherited from the original CleanLab code (and so are their descriptions)

        :param cv_n_folds: int
          This class needs holdout predicted probabilities for every data example
          and if not provided, uses cross-validation to compute them.
          cv_n_folds sets the number of cross-validation folds used to compute
          out-of-sample probabilities for each example in X.

        :param prune_method: str (default: 'prune_by_noise_rate')
          Available options: 'prune_by_class', 'prune_by_noise_rate', or 'both'.
          This str determines the method used for pruning.
          1. 'prune_by_noise_rate': works by removing examples with *high probability* of being mislabeled for every
            non-diagonal in the `prune_counts_matrix` (see pruning.py).
          2. 'prune_by_class': works by removing the examples with *smallest probability* of belonging to their given
            class label for every class.
          3. 'both': Finds the examples satisfying (1) AND (2) and removes their set conjunction.

        :param converge_latent_estimates: bool (Default: False)
          If true, forces numerical consistency of latent estimates. Each is
          estimated independently, but they are related mathematically with closed
          form equivalences. This will iteratively enforce consistency.

        :param pulearning : int (0 or 1, default: None).
            Only works for 2 class datasets. Set to the integer of the class that is perfectly labeled
            (certain no errors in that class).

        :param n_jobs: int (Windows users may see a speed-up with n_jobs = 1).
            Number of processing threads used by
            multiprocessing. Default None sets to the number of processing threads on your CPU. Set this to 1 to REMOVE
            parallel processing (if its causing issues).
        """

        super().__init__(**kwargs)
        self.cv_n_folds = cv_n_folds
        self.prune_method = prune_method
        self.converge_latent_estimates = converge_latent_estimates
        self.pulearning = pulearning
        self.n_jobs = n_jobs
        self.psx_calculation_method = psx_calculation_method



import logging
import random
from typing import Dict

import numpy as np
import torch
from torch.utils.data import TensorDataset

logger = logging.getLogger(__name__)


def get_labels_randomly(
        rule_matches_z: np.ndarray, rule_assignments_t: np.ndarray
) -> np.ndarray:
    """ Calculates sample labels basing on z and t matrices. If several patterns matched, select one randomly """

    if rule_matches_z.shape[1] != rule_assignments_t.shape[0]:
        raise ValueError("Dimensions mismatch!")

    one_hot_labels = rule_matches_z.dot(rule_assignments_t)
    one_hot_labels[one_hot_labels > 0] = 1
    labels = [np.random.choice(np.where(r == 1)[0], 1)[0] for r in one_hot_labels]
    return np.stack(labels, axis=0)


def vocab_and_vectors(filename: str, special_tokens: list) -> (dict, dict, np.ndarray):
    """special tokens have all-zero word vectors"""
    with open(filename, encoding="UTF-8") as in_file:
        parts = in_file.readline().strip().split(" ")
        num_vecs = int(parts[0]) + len(special_tokens)  # + 1
        dim = int(parts[1])

        matrix = np.zeros((num_vecs, dim))
        word_to_id = dict()

        nextword_id = 0
        for token in special_tokens:
            word_to_id[token] = nextword_id
            nextword_id += 1

        for line in in_file:
            parts = line.strip().split(" ")
            word = parts[0]
            if word not in word_to_id:
                emb = [float(v) for v in parts[1:]]
                matrix[nextword_id] = emb
                word_to_id[word] = nextword_id
                nextword_id += 1
    return word_to_id, matrix


def get_embedding_matrix(pretrained_embedding_file: str) -> np.ndarray:
    """ Return matrix with pretrained glove embeddings"""
    with open(pretrained_embedding_file, encoding="UTF-8") as in_file:
        emb_matrix_size = in_file.readline().strip().split(" ")
        embeddings = []
        for line in in_file:
            parts = line.strip().split(" ")
            embeddings.append([float(v) for v in parts[1:]])
        emb_matrix = np.array(embeddings)
        assert emb_matrix.shape[0] == int(emb_matrix_size[0]) and emb_matrix.shape[
            1
        ] == int(emb_matrix_size[1])
    return emb_matrix


def set_device(enable_cuda: bool):
    """ Set where the calculations will be done (cpu or cuda) depending on whether the cuda is available and chosen """
    if enable_cuda and torch.cuda.is_available():
        logger.info("Using GPU")
        return torch.device("cuda")
    else:
        logger.info("Using CPU")
        return torch.device("cpu")


# deprecated
def check_splitting(
        tst_samples: TensorDataset,
        tst_labels: np.ndarray,
        tst_idx: np.ndarray,
        samples: torch.Tensor,
        labels: np.ndarray,
) -> None:
    """ Custom function to check that the splitting into train and test sets fro WSCrossWeigh was done correctly"""

    rnd_tst = np.random.randint(0, tst_samples.tensors[0].shape[0])  # take some random index
    tst_sample = tst_samples.tensors[0][rnd_tst, :]
    tst_idx = tst_idx[rnd_tst]

    tst_label = tst_labels[rnd_tst, :] if len(tst_labels.shape) > 1 else tst_labels[rnd_tst]
    tst_label_true = labels[tst_idx, :] if len(labels.shape) > 1 else tst_labels[rnd_tst]

    if not torch.equal(tst_sample, samples[tst_idx, :]):
        raise RuntimeError(
            "The splitting of original training set into cw train and test sets have been done "
            "incorrectly! A sample does not correspond to one in original dataset"
        )

    if not np.array_equal(tst_label, tst_label_true):
        raise RuntimeError(
            "The splitting of original training set into cw train and test sets have been done "
            "incorrectly! A sample label does not correspond to one in original dataset"
        )


def return_unique(where_to_find: np.ndarray, what_to_find: np.ndarray) -> np.ndarray:
    """ Checks intersections between the 1st and the 2nd arrays and return unique values of the 1st array """
    intersections = np.intersect1d(where_to_find, what_to_find, return_indices=True)[1].tolist()
    return np.delete(where_to_find, intersections)



import numpy as np
import torch
from torch.utils.data import TensorDataset

from knodle.transformation.filter import filter_empty_probabilities, filter_probability_threshold


def test_filter_empty_probabilities():
    input_ids = np.ones((3, 4))
    input_ids[0, 0] = 0
    input_mask = np.ones((3, 4))
    input_mask[1, 1] = 0
    class_probs = np.array([
        [0.5, 0.5],
        [0.3, 0.7],
        [0.0, 0.0]
    ])

    gold_ids = np.ones((2, 4))
    gold_ids[0, 0] = 0
    gold_mask = np.ones((2, 4))
    gold_mask[1, 1] = 0
    gold_probs = np.array([
        [0.5, 0.5],
        [0.3, 0.7]
    ])

    input_dataset = TensorDataset(torch.from_numpy(input_ids), torch.from_numpy(input_mask))

    new_input_dataset, new_probs = filter_empty_probabilities(input_dataset, class_probs)

    assert np.array_equal(new_input_dataset.tensors[0].detach().numpy(), gold_ids)
    assert np.array_equal(new_input_dataset.tensors[1].detach().numpy(), gold_mask)
    assert np.array_equal(new_probs, gold_probs)


def test_filter_probability_threshold():
    input_ids = np.ones((3, 4))
    input_ids[0, 0] = 0
    input_mask = np.ones((3, 4))
    input_mask[1, 1] = 0
    class_probs = np.array([
        [0.5, 0.5],
        [0.3, 0.7],
        [0.0, 0.0]
    ])

    gold_ids = np.ones((1, 4))
    gold_mask = np.ones((1, 4))
    gold_mask[0, 1] = 0
    gold_probs = np.array([
        [0.3, 0.7]
    ])

    input_dataset = TensorDataset(torch.from_numpy(input_ids), torch.from_numpy(input_mask))

    new_input_dataset, new_probs = filter_probability_threshold(input_dataset, class_probs, probability_threshold=0.6)

    assert np.array_equal(new_input_dataset.tensors[0].detach().numpy(), gold_ids)
    assert np.array_equal(new_input_dataset.tensors[1].detach().numpy(), gold_mask)
    assert np.array_equal(new_probs, gold_probs)




from typing import Callable

from knodle.trainer.config import TrainerConfig


class AutoConfig:
    """ The factory class for creating Config classes of training executors
    See See https://medium.com/@geoffreykoh/implementing-the-factory-
    pattern-via-dynamic-registry-and-python-decorators-479fc1537bbe
    """

    registry = {}
    """ Internal registry for available trainers """

    def __init__(self, name, **kwargs):
        self.config = self.create_config(name, **kwargs)

    @classmethod
    def create_config(cls, name: str, **kwargs) -> TrainerConfig:

        exec_class = cls.registry[name]
        executor = exec_class(**kwargs)
        return executor

    @classmethod
    def register(cls, name: str) -> Callable:
        def inner_wrapper(wrapped_class: TrainerConfig) -> Callable:
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper



import numpy as np

from knodle.transformation.labels import label_ids_to_labels


def test_label_ids_to_labels():
    prediction_ids = np.array([0, 1, 1, 0])
    gold_label_ids = np.array([1, 0, 0, 1])

    labels2ids = {
        0: "first",
        1: "second"
    }

    predictions_truth = ["first", "second", "second", "first"]
    gold_labels_truth = ["second", "first", "first", "second"]

    predictions, gold_labels = label_ids_to_labels(prediction_ids, gold_label_ids, labels2ids)

    assert predictions == predictions_truth
    assert gold_labels == gold_labels_truth



from typing import Dict, Union, List

import pandas as pd
import numpy as np


def get_y_statistics(y_gold: Union[List, np.ndarray]) -> pd.DataFrame:
    """Returns a few statistics for a label vector.
    """
    y = pd.Series(y_gold)
    stats_dict = [
        ["num_classes", y.nunique()],
        ["num_samples", y_gold.shape[0]],
        ["classes", y.unique().tolist()],
        ["class_sums", pd.Series(y_gold).value_counts().to_dict()],
        ["class_distribution", (pd.Series(y_gold).value_counts() / y_gold.shape[0]).to_dict()]
    ]
    stats_dict = pd.DataFrame(stats_dict, columns=["statistic", "value"])
    return stats_dict


def get_z_t_statistics(rule_matches_z: np.array, mapping_rules_labels_t: np.array) -> pd.DataFrame:
    """Returns a few statistics for a label vector.
    """
    hits = np.round(rule_matches_z.sum(axis=0) / rule_matches_z.shape[0], 2)
    no_match_samples = np.where(hits == 0)[0].sum() / rule_matches_z.sum()

    avg_hits_per_sample = rule_matches_z.sum() / rule_matches_z.shape[0]

    stats_dict = {
        "num_samples": rule_matches_z.shape[0],
        "num_classes": mapping_rules_labels_t.shape[1],
        "num_weak_labels": rule_matches_z.shape[1],
        "samples_without_match": no_match_samples,
        "avg_hits_per_sample": avg_hits_per_sample
    }

    stats_dict = pd.DataFrame(stats_dict, columns=["statistic", "value"])
    return stats_dict


def get_standard_paper_stats(
        rule_matches_z: np.array, mapping_rules_labels_t: np.array, y_gold: np.ndarray
) -> pd.DataFrame:
    """Computes basic statistics relevant for a paper.
    """
    if mapping_rules_labels_t.shape[1] == 2:
        skewdness = round(y_gold.sum() / y_gold.shape[0], 2)
    else:
        skewdness = None

    stats_dict = [
        ["classes", (y_gold.max() + 1).astype(str)],
        ["train / test samples", f"{rule_matches_z.shape[0]} / {y_gold.shape[0]}"],
        ["rules", rule_matches_z.shape[1]],
        ["avg. rule hits", round(rule_matches_z.sum() / rule_matches_z.shape[0], 2)],
        ["skewdness", skewdness]
    ]
    stats_dict = pd.DataFrame(stats_dict, columns=["statistic", "value"])
    return stats_dict


def combine_multiple_paper_stats(dataset_to_df_dict: Dict) -> pd.DataFrame:
    """Takes a dictionary, having dataset names as keys and DataFrames as values and returns a combined DataFrame.
    """
    columns = []
    values = []

    for dataset, df in dataset_to_df_dict.items():
        columns = df["statistic"].tolist()
        values.append([dataset] + df["value"].tolist())

    stats_df = pd.DataFrame(values, columns=["dataset"] + columns)
    return stats_df



from typing import Dict
import numpy as np


def label_ids_to_labels(predictions: np.ndarray, labels: np.ndarray, ids2labels: Dict) -> [np.ndarray, np.ndarray]:
    """
    Prepare data for string label based evaluation.

    Args:
        predictions: predicted label ids
        labels: gold label ids
        ids2labels: translation dictionary from string labels to label ids
    Returns:
        Two lists of class labels for predictions and gold labels
    """
    predictions_idx = predictions.astype(int).tolist()
    labels_idx = labels.astype(int).tolist()

    predictions = [ids2labels[p] for p in predictions_idx]
    gold_labels = [ids2labels[p] for p in labels_idx]
    return predictions, gold_labels
